# =====================================================================
# mainlib.awk: common AWK functions for CSA.
#
# Copyright (c) 2001,2002,2003 Carlo Strozzi
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# =====================================================================
# $Id: mainlib.awk,v 1.3 2003/05/07 09:11:26 carlo Exp $

# =====================================================================
# Standard C-library workalikes (not always compatible, often quite
# different from their equally-named C counterparts).
# =====================================================================

# =====================================================================
# strdup(3): string strdup(string s, int length)
# =====================================================================

function strdup(s1, len,	s2) {

   if (!len) len = 2					# see strdup(3)
   s2 = sprintf("%*s", (len + length(s1) - 1), s1)
   gsub(/ /, s1, s2)
   return s2
}

# =====================================================================
# ctime(3): string ctime(string specs)
# =====================================================================

function ctime(ts,	format, t) {

   format = "%a %b %d %H:%M:%S %Z %Y"
   if (ts == 0) CMD = "date '+" format "'"
   else
      CMD = "date -d'1970-01-01 UTC + " ts " sec' '+" format "'"

   CMD = "exec " CMD				# Spare processes.
   CMD | getline t
   close(CMD)
   return t
}

# =====================================================================
# stat(2): int stat(string file)
#
# Returns EINVAL if no arg, 0 if file exists (can be a hard/symbolic
# link) and is readable, EISDIR if file is a directory, ENOENT if file
# doesn't exist.
# =====================================================================

function stat(file,	rc) {

   if (file == "") return EINVAL
   if ((rc = (getline < file)) >= 0) { close(file); return 0 }
   if (rc == 2) return EISDIR
   else if (rc == 255) return ENOENT
   else return EIO		      # Unknown code, assume I/O error.
}

# =====================================================================
# creat(2): int creat(string file [, int flag])
#
# 'flag' can be either O_TRUNC or O_APPEND. The file will be left
# open in the specified mode on return.
# =====================================================================

function creat(file,flag,	rc) {

   if (file == "") return EINVAL
   if (flag != O_APPEND) flag = O_TRUNC

   # Check whether we can write to the requested file.
   printf("") >> file
   if ((rc=close(file)) != 0) return EACCES

   # Yes, we can.
   if (flag == O_APPEND) { printf("") >> file; return 0 }
   else printf("") > file

   return 0
}

# =====================================================================
# CSA basic functions.
# =====================================================================

# =====================================================================
# string strip(string s, [boolean middle])
#
# Strip leading and trailing blanks and TABs from the input string.
# Returns the stripped string. If 'middle' is set to '1', then multiple
# middle blanks are squeezed.
# =====================================================================

function strip(s,m) {
  sub(/^[ \t]+/,NULL,s)
  sub(/[ \t]+$/,NULL,s)
  if (m) gsub(/  +/," ",s)
  return s
}

# =====================================================================
# boolean rcassign(string name, string value [, string ofile])
#
# Turn the passed name/value pair into a valid rc(1) assignment in
# the form: name=(value_1 value_2 ... value_n). Value is split on
# octal "\001" to cope with rc(1) lists. The resulting assignment is
# then printed to 'ofile', properly escaped for rc(1).
# =====================================================================

function rcassign(name,value,ofile,		i,a,j) {

  i=j=0
  if (ofile == NULL) ofile = STDOUT

  # Help rc(1) handle non-conventional variable names.
  name = rcescape(name,1)
  if (name == "''") return EINVAL

  printf("%s=(",name) >> ofile

  # Split on "\001" to cope with rc(1) lists.

  if ((i = split(value,a,"\001"))) { 
     for (j=1; j<=i; j++) printf(" %s", rcescape(a[j],1)) >> ofile
  }

  print ")" >> ofile
}

# =====================================================================
# string rcescape(string s [, boolean set])
#
# Escape single-quotes in 'string' by doubling them as required by
# rc(1). The resulting string is then enclosed in single-quotes. If
# 'string' is null, then '()' is returned, unless the 'set' argument is
# set to '1', in which case the empty string assignment '' is returned.
# See rc(1) for more info.
# =====================================================================

function rcescape(string,set) {

  if (string == NULL) {
     if (set) return "''"
     else return "()"
  }

  gsub("'", "''", string)
  return "'" string "'"
}

# =====================================================================
# nil rcmsg(string msg [, string ofile])
#
# Takes a string with all the parameters necessary for the invoking
# rc(1) script to call the underlying csaPrintMsg() library function.
# Individual parameters are extracted by splitting 'string' on octal
# value "\001".
# =====================================================================

function rcmsg(msg, ofile,		i,a) {

  if (ofile == "") ofile = STDOUT

  i = split(msg, a, "\001")

  if (a[1] !~ /^[0-9]+$/ || a[2] !~ /^[IWE]$/) return EINVAL

  printf("~ $CSA_WBMSG () && {CSA_WBMSG=(%s %c", a[1], a[2]) >> ofile

  if (a[3] != "") printf(" %s)}\n", rcescape(a[3])) >> ofile
}

# =====================================================================
# boolean isemail(string s)
#
# Checks that 'string' is a valid E-Mail address. Returns non-zero if
# it isn't. Multiple addresses can be specified in 'string', by
# saparating them with a comma ','. Blanks and/or tabs surrounding the
# address separator are silently ignored.
# =====================================================================

function isemail(string,	addr1,addr2) {

  gsub(/[ \t]*,[ \t]*/, ",", string)	# Strip empty separators.
  if(string == NULL) return EINVAL

  split(string, addr1, ",")

  for (addr2 in addr1) {
      if (addr1[addr2] !~ /^[-_=.a-zA-Z0-9]+@[-.a-zA-Z0-9]+$/)
		return EINVAL
  }
}

# =====================================================================
# string unesctbl(string s)
#
# Un-escape NoSQL special sequences \t, \n and \\
# Note: global variables, like RSTART and RLENGTH cannot be localised
# with their original names, or things won't work.
# =====================================================================

function unesctbl(s,		S,rstart,rlength) {

   while (match(s,/\\+t/)) {

      rstart=RSTART
      rlength=RLENGTH

      if (!(rlength % 2))
	 S = S substr(s,1,rstart+rlength-3) "\t"
      else
	 S = S substr(s,1,rstart+rlength-1)

      s = substr(s,rstart+rlength)
   }

   s = S s
   S = ""

   while (match(s,/\\+n/)) {

      rstart=RSTART
      rlength=RLENGTH

      if (!(rlength % 2))
	 S = S substr(s,1,rstart+rlength-3) "\n"
      else
	 S = S substr(s,1,rstart+rlength-1)

      s = substr(s,rstart+rlength)
   }

   s = S s
   S = ""

   while (match(s,/[\\][\\]/)) {

      rstart=RSTART; rlength=RLENGTH

      S = S substr(s,1,rstart)
      s = substr(s,rstart+2)
   }

   return S s
}

# =====================================================================
# string escreg(string re)
#
# Escape special characters in a string that is to be used as an AWK
# regular expression.
# =====================================================================

function escreg(re) {

  # Escape backslashes first ...
  gsub(/\\/,"\\\\\\", re)

  # ... then all other special characters.
  gsub(/\$/, "\\$", re)
  gsub(/\(/, "\\(", re)
  gsub(/\)/, "\\)", re)
  gsub(/\[/, "\\[", re)
  gsub(/\]/, "\\]", re)
  gsub(/\|/, "\\|", re)
  gsub(/\^/, "\\^", re)
  gsub(/\*/, "\\*", re)
  gsub(/\?/, "\\?", re)
  gsub(/\./, "\\.", re)

  return re
}

# =====================================================================
# string escape(string s)
#
# Escape AWK special characters ", \ and \n in a string that is to be
# eval'd again by AWK.
# =====================================================================

function escape(s) {

  # Escape backslashes first ...
  gsub(/\\/,"\\\\\\", s)

  # ... then double-quotes and newlines.
  gsub(/"/, "\\\"", s)
  gsub(/\n/, "\\n", s)

  return "\"" s "\""
}

# =====================================================================
# boolean awkassign(string name, array value \
#			[, boolean null, boolean begin, string ofile])
#
# Turn the 'value' array into a valid AWK array assignment statement.
# The No. of elements
# is expected in value[0]. The resulting assignment is then printed to
# stdout, properly escaped for AWK. Unless 'null' is non-zero, empty
# input values will result in a "delete" statement on output.
# If 'begin' is non-zero, the resulting assignment is done inside a
# "BEGIN{}" block.
# =====================================================================

function awkassign(name,value,null,begin,ofile,		i) {

  i=0
  if (ofile == NULL) ofile = STDOUT

  # Check for legal name.
  if (name !~ /^[A-Za-z][A-Za-z0-9_]*$/) return EINVAL

  if (begin) print "BEGIN {" >> ofile

  if (value[0] <= 0) {			# Handle special case.
     if (!null) printf("delete %s\n", name) >> ofile
     else printf("%s=\"\"", name) >> ofile
     return 0
  }

  printf("%s[0]=%d\n", name, value[0]) >> ofile

  while (++i <= value[0]) {
     printf("%s[%d]=%s\n", name, i, escape(value[i])) >> ofile
  }

  if (begin) print "}" >> ofile
}

# =====================================================================
# string b64decode(string value)
#
# Take a base64-encoded string and decode it to its original value. 
# Being mawk(1) not 8-bit clean, the original unencode string should
# not contain values that fall outside the LOCALE charset.
# =====================================================================

function b64decode(value,		S,n,k,f,i,s,b,z,c) {

  n["/"] = 63
  n["+"] = 62

  for (k=26;k;) {
      n[sprintf(f="%c",--k+65)]=k
      if (k<10) n[sprintf(f,k+48)]=k+52
      n[sprintf(f,97+k)]=26+k
  }

  for (i=1;i<=length(value);) {
      if ((s=substr(value,i++,1)) in n){
         c=n[s]
         if (1==k%=4) {
	    b=z*4+(c-(z=c%16))/16
         }
         else if (!k) z=c
         else if (k==2) {
           b=z*16+(c-(z=c%4))/4
         }
         else if (k==3) b=z*64+c

         if (k++) S = S sprintf(f,b)
      }
  }

  return S
}

# =====================================================================
# nil authset(string auth)
#
# Handle basic HTTP authorization.
# =====================================================================

function authset(s,ofile,		a,S,err) {

   if (ofile == "") ofile = STDOUT

   if (s == "") return				# ignore null requests.

   S = s					# Copy to workarea.

   if (!sub(/^[Bb][Aa][Ss][Ii][Cc] +/,"",S)) err=1	# Basic auth only
   if (S !~ /^[a-zA-Z0-9]+=*[\r]?$/) err=1		# Base64 only

   if (err) {
      # Clear all auth vars on error.
      print "HTTP_AUTHORIZATION=();CSA_AUTH_PBC=()" >> ofile
      print "CSA_AUTH_USER=();CSA_AUTH_PW=()" >> ofile
      return
   }

   split(b64decode(S),a,":")

   # For user-id and password to be valid, they must contain only
   # allowed characters. Note that the password field may be null,
   # and that this is acceptable.

   if (a[1] ~ /^[a-z][a-z][-_a-z0-9@.]*$/ && \
			a[2] ~ /^[-+_a-zA-Z0-9@.!]*$/) {

      printf("CSA_AUTH_USER='%s'\n", a[1]) >> ofile
      printf("CSA_AUTH_PBC=%s/%s\n",\
			substr(a[1],1,1),substr(a[1],2,1)) >> ofile
      rcassign("CSA_AUTH_PW", a[2], ofile)

      # The application-level token must be passed to the application,
      # to be re-sent to the client with every response if necessary.
      # Note the re-use of the same variable that would be set by the
      # server in case of HTTP authorization.

      printf("HTTP_AUTHORIZATION='%s'\n", s) >> ofile
   }

   else {
      # If anything is wrong, clear all auth vars.
      print "HTTP_AUTHORIZATION=();CSA_AUTH_PBC=()" >> ofile
      print "CSA_AUTH_USER=();CSA_AUTH_PW=()" >> ofile
   }

   # Server-based authentication, if used, *MUST* override the one
   # attained by all other means, and the application-level token
   # must be cleared. If the server token does not match the allowed
   # pattern, then it is cleared too.

   if (ENVIRON["REMOTE_USER"] ~ /^[a-z][a-z][-_a-z0-9@.]*$/) {
      printf("CSA_AUTH_USER='%s'\n", ENVIRON["REMOTE_USER"]) >> ofile
      printf("CSA_AUTH_PBC=%s/%s\n",substr(ENVIRON["REMOTE_USER"],1,1),\
			substr(ENVIRON["REMOTE_USER"],2,1)) >> ofile
      print "CSA_AUTH_OK=1;CSA_AUTH_PW=();HTTP_AUTHORIZATION=()" >> ofile
   }
   else print "REMOTE_USER=()" >> ofile
}

# =====================================================================
# Main program settings.
# =====================================================================

BEGIN {

  # The CMD global variable is used to hold system commands. It is
  # a global name because AWK can have only one system command opened
  # at a time, and the current command needs to be closed before a new
  # one can be opened.

  NULL = RESULT = CMD = ""

  # Linux-style symbolic error codes (see errno(3) and <errno.h>).
  ENOENT	=   2
  EIO		=   5
  EACCES	=  13
  EISDIR	=  21
  EINVAL	=  22
  ENOMSG	=  42
  EMSGSIZE	=  90

  # Various flags.
  O_TRUNC=0; O_APPEND=1

  # Get local settings.
  CSAINST = ENVIRON["CSA_INSTALL"]
  STDOUT = ENVIRON["CSA_STDOUT"]
  STDERR = ENVIRON["CSA_STDERR"]

  # Set default values if necessary.
  if (CSAINST == NULL) CSAINST = "/usr/local/csa"
  if (STDOUT == NULL) STDOUT = "/dev/stdout"
  if (STDERR == NULL) STDERR = "/dev/stderr"
}

# =====================================================================
# End of mainlib.awk
# =====================================================================
